#!/usr/bin/env python
# =======================================
# Python Physics (v1.0)
# -------------------------
# By M. Pêcheux (June 2018)
# Under Creative Commons BY-NC-SA License
# =======================================
#
# 'Collisions' simulation: spawns some
# particles in a top-view box that collide
# together and bounce off the borders.

from resources import *
import getopt

## UTIL FUNCTIONS
## ==============
def usage():
    print('======================================')
    print('Physics: Collisions (v1.0)')
    print('-------------------------')
    print('By M. Pêcheux (June 2018)')
    print('======================================\n')

    print('Usage: python ' + sys.argv[0])
    print('with the following possible options:\n')
    print('-h, --help\t\tto display this help')
    print('-t=T\t\t\tto specify the end time of the animation')
    print('-n, --nb_particles=N\tto specify the number of particles in the box')
    print('--friction=R\t\tto specify the friction coefficient')
    print('--elasticity=R\t\tto specify the elasticity coefficient')
    print('-s, --show\t\tto show the result (instead of saving it to .mp4)')
    print('')
    print('(Other options can be set directly in the Python file...)')
## ==============

## MODEL-DEPENDENT FUNCTIONS
## =========================
def update(data, flags):
    # apply velocity
    data[:,0] += data[:,2]
    data[:,1] += data[:,3]

    nb_parts = data.shape[0]
    for n in range(nb_parts):
        # apply friction
        data[n,2] *= flags['coeff_friction']
        data[n,3] *= flags['coeff_friction']
        if abs(data[n,2]) < 0.01:
            data[n,2] = 0.
        if abs(data[n,3]) < 0.01:
            data[n,3] = 0.

        # get particle data
        x, y, vx, vy, s, r = data[n,:]
        # bouncing
        if x > BOX_SIZE - flags['box_bounds'] - r:
            data[n,0] = BOX_SIZE - flags['box_bounds'] - r
            data[n,2] = -vx

        elif x < flags['box_bounds'] + r:
            data[n,0] = flags['box_bounds'] + r
            data[n,2] = -vx

        if y > BOX_SIZE - flags['box_bounds'] - r:
            data[n,1] = BOX_SIZE - flags['box_bounds'] - r
            data[n,3] = -vy

        elif y < flags['box_bounds'] + r:
            data[n,1] = flags['box_bounds'] + r
            data[n,3] = -vy

        # collisions
        for other in range(n+1, nb_parts):
            dist = dist_numpy(data, n, other)
            radius_sum = r + data[other,5]
            if dist <= radius_sum:
                # get other particle's data
                xo, yo, vxo, vyo = data[other,:4]
                # we get the temporary components of the vector that is
                # normal to the collision (not normalized)
                tmp_nx, tmp_ny = x - xo, y - yo

                # normalized components
                norm = np.sqrt(tmp_nx ** 2 + tmp_ny ** 2)
                nx, ny = tmp_nx / norm, tmp_ny / norm

                # tangent normalized vector (pi/2 rotation in trigonometric direction)
                gx, gy = -ny, nx

                # if balls are overlapping, we reposition them next to each other
                # overlapping amount
                delta = radius_sum - dist
                data[n,0], data[n,1] = (x + nx * delta / 2, y + ny * delta / 2)
                data[other,0], data[other,1] = (xo - nx * delta / 2, yo - ny * delta / 2)

                # velocities in the new orthonormal base (n, g)
                # base transformation matrix
                # self velocity
                vel_n, vel_g = nx * vx + ny * vy, gx * vx + gy * vy
                # other ball's velocity
                other_vel_n, other_vel_g = nx * vxo + ny * vyo, gx * vxo + gy * vyo

                # collision: normal components are inversed, tangent ones are untouched
                vel_n, other_vel_n = other_vel_n, vel_n

                # back to the normal base, we get the new velocity vectors
                data[n,2] = nx * vel_n + gx * vel_g
                data[n,3] = ny * vel_n + gy * vel_g
                data[other,2] = nx * other_vel_n + gx * other_vel_g
                data[other,3] = ny * other_vel_n + gy * other_vel_g

                # apply elasticity coeff
                data[n,2] *= flags['coeff_elasticity']
                data[n,3] *= flags['coeff_elasticity']
                data[other,2] *= flags['coeff_elasticity']
                data[other,3] *= flags['coeff_elasticity']
## =========================

## MAIN FUNCTIONS
## =========================
def define_flags():
    return {
        'nb_particles': 15,
        'box_bounds': 0.,
        'coeff_friction': .975,
        'coeff_elasticity': .9,
        'anim_length': 100,
        'anim_export': True,
    }

def main(flags):
    # a flags['nb_particles'] * 4 array:
    # - each line is a particle
    # - columns are: x, y, vx, vy, size, radius
    np.random.seed()

    # set up random particles
    particles_data = np.empty((flags['nb_particles'], 6))
    for i in range(flags['nb_particles']):
        p = Particle()
        particles_data[i] = np.array([p.x, p.y, p.vx, p.vy, p.size, p.radius])

    # set up figure
    fig = plt.figure(figsize=(5, 5), dpi=DPI)
    ax = plt.Axes(fig, [0.05, 0.05, 0.9, 0.9])
    ax.set_axis_off()
    ax.set_xlim([0, BOX_SIZE])
    ax.set_ylim([0, BOX_SIZE])
    fig.add_axes(ax)
    plt.hlines(y=0, xmin=0, xmax=BOX_SIZE)
    plt.hlines(y=BOX_SIZE, xmin=0, xmax=BOX_SIZE)
    plt.vlines(x=0, ymin=0, ymax=BOX_SIZE)
    plt.vlines(x=BOX_SIZE, ymin=0, ymax=BOX_SIZE)

    ims = []
    for i in range(flags['anim_length']):
        # compute iteration
        update(particles_data, flags)
        # plot iteration
        ims.append([plt.scatter(
            particles_data[:,0],
            particles_data[:,1],
            s=particles_data[:,4] * SIZE_RATIO,
            c='k'
        ),])

    # create and export animation
    Writer = animation.writers['ffmpeg']
    writer = Writer(fps=15, metadata=dict(artist='Me'), bitrate=1800)

    im_ani = animation.ArtistAnimation(
        fig, ims, interval=1, blit=False
    )
    if flags['anim_export']:
    	im_ani.save('collisions.mp4', writer=writer)
    else:
    	plt.show()

if __name__ == '__main__':
    flags = define_flags()
    # parse command line arguments
    try:
        opts, args = getopt.getopt(
            sys.argv[1:], 't:n:hs',
            ['help', 'nb_particles=', 'friction=', 'elasticity=', 'show']
        )
    except getopt.GetoptError as err:
        # print help information and exit:
        print(err)
        usage()
        sys.exit(2)

    for o, a in opts:
        if o == '-t':
            flags['anim_length'] = int(a)
        elif o in ('-s', '--show'):
            flags['anim_export'] = False
        elif o in ('-n', '--nb_particles'):
            flags['nb_particles'] = int(a)
        elif o == '--friction':
            flags['coeff_friction'] = float(a)
        elif o == '--elasticity':
            flags['coeff_elasticity'] = float(a)
        elif o in ('-h', '--help'):
            usage()
            sys.exit()
        else:
            assert False, 'unhandled option'

    main(flags)
