#!/usr/bin/env python
# =======================================
# Python Physics (v1.0)
# -------------------------
# By M. Pêcheux (June 2018)
# Under Creative Commons BY-NC-SA License
# =======================================
#
# 'Pendulum' simulation: spawns some
# pendulums with one or more hinges (ie
# handle points) that turn around.

from resources import *
import getopt

## UTIL FUNCTIONS
## ==============
def usage():
    print('======================================')
    print('Physics: Pendulums (v1.0)')
    print('-------------------------')
    print('By M. Pêcheux (June 2018)')
    print('======================================\n')

    print('Usage: python ' + sys.argv[0])
    print('with the following possible options:\n')
    print('-h, --help\t\tto display this help')
    print('-t=T\t\t\tto specify the end time of the animation')
    print('--friction=R\t\tto specify the friction coefficient')
    print('-s, --show\t\tto show the result (instead of saving it to .mp4)')
    print('')
    print('(Other options can be set directly in the Python file...)')
## ==============

## MODEL-DEPENDENT FUNCTIONS
## =========================
def update(pendulums, t, flags):
    for pendulum in pendulums:
        pendulum.update(t, flags)
        pendulum.update_visual()
## =========================

## MAIN FUNCTIONS
## =========================
def define_flags():
    return {
        'coeff_friction': .975,
        'anim_length': 100,
        'anim_export': True,
    }

# init function
def init(lines, pendulums):
    line_id = 0
    for p in pendulums:
        for h in range(1, len(p.hinges)):
            lines[line_id].set_data(
                [p.hinges[h-1][0], p.hinges[h][0]],
                [p.hinges[h-1][1], p.hinges[h][1]]
            )
            line_id += 1
    return lines,

# anim function
def anim(i, lines, pendulums, flags):
    update(pendulums, i, flags)
    line_id = 0
    for p in pendulums:
        for h in range(1, len(p.hinges)):
            lines[line_id].set_data(
                [p.hinges[h][0], p.hinges[h-1][0]],
                [p.hinges[h][1], p.hinges[h-1][1]]
            )
            line_id += 1
    return lines,

def main(flags):
    # set up figure
    fig = plt.figure(figsize=(5, 5), dpi=DPI)
    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.set_axis_off()
    fig.add_axes(ax)
    plt.hlines(y=0, xmin=0, xmax=BOX_SIZE)
    plt.hlines(y=BOX_SIZE, xmin=0, xmax=BOX_SIZE)
    plt.vlines(x=0, ymin=0, ymax=BOX_SIZE)
    plt.vlines(x=BOX_SIZE, ymin=0, ymax=BOX_SIZE)

    # set up pendulums
    pendulums = [
        Pendulum(hinges=[
            [0.25*BOX_SIZE, BOX_SIZE],
            [0.15*BOX_SIZE, 0.15*BOX_SIZE],
        ]),
        Pendulum(hinges=[
            [0.5*BOX_SIZE, 0.85*BOX_SIZE],
            [0.6*BOX_SIZE, 0.65*BOX_SIZE],
            [0.45*BOX_SIZE, 0.45*BOX_SIZE],
            [0.5*BOX_SIZE, 0.35*BOX_SIZE],
        ]),
        Pendulum(hinges=[
            [0.75*BOX_SIZE, 0.35*BOX_SIZE],
            [0.77*BOX_SIZE, 0.45*BOX_SIZE],
        ]),
    ]

    lines = []
    for pendulum in pendulums:
        for hingejoint in pendulum.hingejoints:
            ax.add_artist(hingejoint)

        for h in range(1, len(pendulum.hinges)):
            line, = plt.plot(
                [pendulum.hinges[h][0], pendulum.hinges[h-1][0]],
                [pendulum.hinges[h][1], pendulum.hinges[h-1][1]],
                c='k'
            )
            lines.append(line)

    # create and export animation
    im_ani = animation.FuncAnimation(
    		fig, anim, flags['anim_length'],
    		init_func=lambda: init(lines, pendulums),
            fargs=(lines, pendulums, flags),
    		interval=100, blit=False
    	)
    if flags['anim_export']:
    	im_ani.save('pendulum.mp4')
    else:
    	plt.show()

if __name__ == '__main__':
    flags = define_flags()
    # parse command line arguments
    try:
        opts, args = getopt.getopt(
            sys.argv[1:], 't:n:hs',
            ['help', 'nb_particles=', 'friction=', 'show']
        )
    except getopt.GetoptError as err:
        # print help information and exit:
        print(err)
        usage()
        sys.exit(2)

    for o, a in opts:
        if o == '-t':
            flags['anim_length'] = int(a)
        elif o in ('-s', '--show'):
            flags['anim_export'] = False
        elif o in ('-n', '--nb_particles'):
            flags['nb_particles'] = int(a)
        elif o == '--friction':
            flags['coeff_friction'] = float(a)
        elif o in ('-h', '--help'):
            usage()
            sys.exit()
        else:
            assert False, 'unhandled option'

    main(flags)
