#!/usr/bin/env python
# =======================================
# Python Physics (v1.0)
# -------------------------
# By M. Pêcheux (June 2018)
# Under Creative Commons BY-NC-SA License
# =======================================
#
# Resource file that defines util classes
# and methods for physics simulations.

import sys
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from matplotlib.patches import Circle

ANIM_EXPORT = True

DPI = 150.

BOX_SIZE = 100
PARTICLE_SIZE = 2
SIZE_RATIO = 10000 * (np.pi ** 2)

G = 9.81

def dist_lists(p1, p2):
    x1, y1 = p1
    x2, y2 = p2
    return np.sqrt((x1-x2) ** 2 + (y1-y2) ** 2)

def dist_numpy(data, n1, n2):
    return np.sqrt((data[n1,0] - data[n2,0]) ** 2 + (data[n1,1] - data[n2,1]) ** 2)

def pts_angle(p1, p2, direction):
    x1, y1 = p1
    x2, y2 = p2
    n1, n2 = dist_lists(p1, p2), dist_lists(direction, [0., 0.])
    if n1 == 0.:
        n1 = 1.
    if n2 == 0.:
        n2 = 1.
    d = np.dot([x2-x1, y2-y1], direction) / (n1*n2)
    a = np.arccos(d)
    cross = np.cross([x2-x1, y2-y1], direction)
    if (np.dot(-1., cross) < 0):
        a = -a;
    if y2 > y1:
        a += np.pi
    return a

class Particle(object):
    """Simple physics object represented by a circle with a position,
    a speed and a size"""

    def __init__(self, x=None, y=None, vx=None, vy=None, radius=None,
        defaults=None):
        if defaults is not None:
            if 'pos' in defaults:
                self.default_pos = defaults['pos']
            if 'speed' in defaults:
                self.default_speed = defaults['speed']
            if 'radius' in defaults:
                self.default_radius = defaults['radius']

        dft_pos = self.default_pos()
        dft_speed = self.default_speed()

        self.x = x if x is not None else dft_pos[0]
        self.y = y if y is not None else dft_pos[1]
        self.vx = vx if vx is not None else dft_speed[0]
        self.vy = vy if vy is not None else dft_speed[1]

        self.radius = radius if radius is not None else self.default_radius()
        self.radius = self.radius / BOX_SIZE * PARTICLE_SIZE
        self.size = np.pi * (self.radius) ** 2

        self.radius *= BOX_SIZE
        self.vx *= BOX_SIZE / 50.
        self.vy *= BOX_SIZE / 50.

    @staticmethod
    def default_pos():
        return (np.random.random() * BOX_SIZE, np.random.random() * BOX_SIZE)

    @staticmethod
    def default_speed():
        return ((np.random.random() - 0.5) * 3, (np.random.random() - 0.5) * 3)

    @staticmethod
    def default_radius():
        return 2. + (np.random.random() - 0.5) * 2.

class Pendulum(object):
    """Basic pendulum with one or more hinges (ie handle points)"""

    def __init__(self, radius=None, hinges=None):
        self.radius = radius if radius is not None else 4.
        self.radius = self.radius / BOX_SIZE
        self.size = np.pi * (self.radius) ** 2

        self.radius *= BOX_SIZE

        if hinges is None:
            self.hinges = [[0.5*BOX_SIZE, BOX_SIZE], [0.25*BOX_SIZE, 0.5*BOX_SIZE]]
        else:
            self.hinges = hinges
        if len(self.hinges) < 2:
            print('Cannot create a pendulum with only one hinge! Specify a second point.')
            sys.exit(2)

        self.lengths = []
        self.angles = []
        self.start_angles = []
        direction = [0., self.hinges[1][1]-self.hinges[0][1]]
        angles_sum = 0.
        for h in range(1, len(self.hinges)):
            self.lengths.append(dist_lists(self.hinges[h-1], self.hinges[h]))
            a = pts_angle(self.hinges[h-1], self.hinges[h], direction)
            self.angles.append(a)
            angles_sum += a
            self.start_angles.append(angles_sum)
            direction = [self.hinges[h][0]-self.hinges[h-1][0], self.hinges[h][1]-self.hinges[h-1][1]]

        self.hingejoints = []
        for h in range(len(self.hinges)):
            if h == len(self.hinges) - 1:
                self.hingejoints.append(Circle(self.hinges[h], self.radius, facecolor='k'))
            if h == 0:
                self.hingejoints.append(Circle(self.hinges[h], 1.25, facecolor='k'))
            else:
                self.hingejoints.append(Circle(self.hinges[h], 0.9, facecolor='k'))

        self.velocity = 1.
        self.update(0)

    def update(self, t, flags={}):
        for i in range(1, len(self.hinges)):
            self.angles[i-1] = self.start_angles[i-1] \
                * np.cos(t/3. * np.sqrt(G/self.lengths[i-1])) * self.velocity \
                - np.pi/2.
            self.hinges[i][0] = np.cos(self.angles[i-1]) * self.lengths[i-1] \
                                + self.hinges[i-1][0]
            self.hinges[i][1] = np.sin(self.angles[i-1]) * self.lengths[i-1] \
                                + self.hinges[i-1][1]

        if 'coeff_friction' in flags:
            self.velocity *= flags['coeff_friction']

    def update_visual(self):
        for i in range(len(self.hinges)):
            self.hingejoints[i].center = self.hinges[i][0], self.hinges[i][1]
